""" Functions that read efficiently files stored on disk """

import os
import re
from typing import Iterator, Optional

import numpy as np
from tqdm import tqdm as tq


def read_shapes_local(local_path: str, reg_exp_pattern: str = r".+\.npy") -> Iterator[np.ndarray]:
    """
    Iterate over shapes saved on disk.

    Parameters
    ----------
    local_embeddings_path : str
        Path on local disk of the embedding in numpy format.

    Returns
    -------
    shape_iterator : Iterator[np.ndarray]
        An iterator over shapes.
    """
    reg_exp = re.compile(reg_exp_pattern)
    filenames = os.walk(local_path).__next__()[2]
    filenames = [filename for filename in filenames if reg_exp.match(filename)]
    filenames.sort()
    for file_name in filenames:
        # https://stackoverflow.com/a/48592009
        shape = np.load(f"{local_path}/{file_name}", mmap_mode="r").shape
        yield shape


def read_embeddings_local(
    local_embeddings_path: str, batch_size: Optional[int] = None, verbose=True, reg_exp_pattern: str = r".+\.npy",
) -> Iterator[np.ndarray]:
    """
    Iterate over embeddings arrays saved on disk.
    It is possible to iterate over batchs of files and yield stacked embeddings arrays.

    The implementation adopted here is chosen for memory concern: it is very important
    for autofaiss to avoid using more memory than necessary.
    In particular, for the faiss training, a large embedding array is nessary.
    It is not possible to save it twice in memory.
    This implementation pre-allocate an array of size batch_size and keep it updated during the iteration over files.
    The maximum memory usage is batch_size * dim * 4

    Parameters
    ----------
    local_embeddings_path : str
        Path on local disk of the embedding in numpy format.
    batch_size : int (default None)
        Outputs a maximum of batch_size vectors, the default is the size of the first file
        This parameter is useful when working with many small files.
    verbose : bool
        Print detailed informations if set to True

    Returns
    -------
    embeddings_iterator : Iterator[np.ndarray]
        An iterator over batchs of stacked embedding arrays.
    """
    try:
        first_vector_count, dim = next(read_shapes_local(local_embeddings_path, reg_exp_pattern))
    except StopIteration as err:
        raise Exception("no files to read from") from err

    if batch_size is None:
        batch_size = first_vector_count

    reg_exp = re.compile(reg_exp_pattern)

    filenames = os.walk(local_embeddings_path).__next__()[2]
    filenames = [filename for filename in filenames if reg_exp.match(filename)]
    filenames.sort()
    embeddings_batch = None
    nb_emb_in_batch = 0

    iterator = filenames
    if verbose:
        iterator = tq(list(iterator))

    for file_name in iterator:
        emb = np.load(f"{local_embeddings_path}/{file_name}", mmap_mode="r")
        vec_size = emb.shape[0]
        current_emb_index = 0
        while True:
            left_in_emb = vec_size - current_emb_index
            remaining_to_add = max(batch_size - nb_emb_in_batch, 0)
            adding = min(remaining_to_add, left_in_emb)
            additional = max(left_in_emb - adding, 0)
            if embeddings_batch is None:
                embeddings_batch = np.empty((batch_size, dim), "float32")
            embeddings_batch[nb_emb_in_batch : (nb_emb_in_batch + adding), :] = emb[
                current_emb_index : (current_emb_index + adding), :
            ]
            nb_emb_in_batch += adding
            current_emb_index += adding
            if nb_emb_in_batch == batch_size:
                yield embeddings_batch
                nb_emb_in_batch = 0
                embeddings_batch = None
            if additional == 0:
                break

    if nb_emb_in_batch > 0 and embeddings_batch is not None:
        yield embeddings_batch[:nb_emb_in_batch]



"""
File that contains the descriptions of the different indices features.
"""

from enum import Enum


class IndexBlock(Enum):
    FLAT = 0
    IVF = 1
    IVF_HNSW = 2
    HNSW = 3
    PQ = 4
    OPQ = 5
    PAD = 6


class TunableParam(Enum):
    EFSEARCH = 0
    NPROBE = 1
    HT = 2


INDEX_DESCRIPTION_BLOCKS = {
    IndexBlock.FLAT: """Flat Index (Simple exhaustive search)
    All the vectors are stored without compression in a 2D array.
    The search is exhaustive and gives exact results, but it can be slow when there are
    many vectors.""",
    IndexBlock.IVF: """IVF (Inverted File Index):
    The vector space is divided into several groups that are represented by a cluster vector.
    The algorithm first looks for the closest cluster vectors by iterating over them one by one.
    Then, The IVF mapping is used to read the vectors in the selected groups. The algorithm reads
    each of them and select the k nearest neighbors.
    -> More info: https://hal.inria.fr/inria-00514462/PDF/jegou_pq_postprint.pdf""",
    IndexBlock.IVF_HNSW: """IVF + HNSW (Inverted File Index + Hierarchical Navigable Small World index):
    The vector space is divided into several groups that are represented by a cluster vector. 
    The algorithm first looks for the closest cluster vectors using an HNSW index (instead of 
    iterating over them one by one in the version without HNSW). Then, The IVF mapping is used
    to read the vectors in the selected groups. The algorithm reads each of them and selects
    the k nearest neighbors.
    -> More info: https://hal.inria.fr/inria-00514462/PDF/jegou_pq_postprint.pdf""",
    IndexBlock.PQ: """PQ (Product Quantization):
    The vectors are compressed using product quantization.
    Some clever optimizations are implemented to compute distances in the compressed space.
    -> More info: https://hal.inria.fr/inria-00514462/PDF/jegou_pq_postprint.pdf""",
    IndexBlock.OPQ: """OPQ (Optimized Product Quantization):
    The vectors are projected using a rotation matrix. The matrix is trained to minimize
    the average compression error.
    -> More info: http://kaiminghe.com/publications/pami13opq.pdf""",
    IndexBlock.HNSW: """HNSW (Hierarchical Navigable Small World):
    All the vectors are stored in a 2D array without compression, and
    a bidirectional graph is built on the top of the vectors to enable a fast search.
    -> More info: https://arxiv.org/ftp/arxiv/papers/1603/1603.09320.pdf""",
    IndexBlock.PAD: """PAD (Padding):
    Preprocessing operations where a padding with 0s is added to the end of the vectors.""",
}


TUNABLE_PARAMETERS_DESCRIPTION_BLOCKS = {
    TunableParam.NPROBE: """    - nprobe: The number of vector groups to explore, the search time is proportional
      to this value. If nprobe is high, the recall will also be high.
      -> More info: https://hal.inria.fr/inria-00514462/PDF/jegou_pq_postprint.pdf""",
    TunableParam.EFSEARCH: """    - efsearch: The number of times a greedy search is done in the HNSW graph. The results
      of the different greedy searches are combined to get a more precise result.
      The search time is proportional to this value. If efsearch is high, the recall will be high.
      -> More info: https://arxiv.org/ftp/arxiv/papers/1603/1603.09320.pdf       """,
    TunableParam.HT: """    - ht (Hamming threshold): A threshold value to approximate the vector distances using the hamming distance.
      Computing the Hamming distances between two vectors is much faster than computing the real distance,
      but it is also less precise. We found out that it was better to deactivate this parameter as it
      decreases the recall score over speed improvements (default value is 2048)
      -> More info: https://arxiv.org/pdf/1609.01882.pdf""",
}



""" test utils functions """
# pylint: disable= invalid-name

import numpy as np
import pytest

from autofaiss.utils.array_functions import multi_array_split


def test_multi_array_split():
    """test multi_array_split fct number 1"""
    assert len(list(multi_array_split([np.zeros((123, 2)), np.zeros((123, 5))], 41))) == 41


@pytest.mark.parametrize("seed", list(range(1, 10)))
def test_multi_array_split_2(seed):
    """test multi_array_split fct number 2"""

    np.random.seed(seed)
    length = np.random.randint(1, 100)
    nb_chunk = np.random.randint(1, length + 1)
    dim1 = np.random.randint(10)
    dim2 = np.random.randint(10)

    a = np.random.randint(0, 10000, (length, dim1))
    b = np.random.randint(0, 10000, (length, dim2))

    c = list(multi_array_split([a, b], nb_chunk))

    a2 = np.concatenate([x[0] for x in c])
    b2 = np.concatenate([x[1] for x in c])

    assert np.all(a == a2)
    assert np.all(b == b2)



""" useful functions to apply on an index """

import os
import time
from itertools import chain, repeat
from pathlib import Path
from typing import Dict, Optional, Union
import random

import faiss
import numpy as np
from autofaiss.utils.os_tools import run_command


def get_index_size(index: faiss.Index) -> int:
    """Returns the size in RAM of a given index"""

    index_file = "/tmp/tmp_index" + str(random.randrange(100000)) + ".idx"

    if os.path.exists(index_file):
        os.remove(index_file)

    faiss.write_index(index, index_file)

    size_in_bytes = Path(index_file).stat().st_size
    os.remove(index_file)

    return size_in_bytes


def speed_test_ms_per_query(
    index: faiss.Index, query: Optional[np.ndarray] = None, ksearch: int = 40, timout_s: Union[float, int] = 5.0
) -> float:
    """Evaluate the average speed in milliseconds of the index without using batch"""

    nb_samples = 2_000

    if query is None:
        query = np.random.rand(nb_samples, index.d).astype("float32")

    count = 0
    nb_repeat = 1 + (nb_samples - 1) // query.shape[0]

    start_time = time.perf_counter()

    for one_query in chain.from_iterable(repeat(query, nb_repeat)):

        _, _ = index.search(np.expand_dims(one_query, 0), ksearch)

        count += 1

        if time.perf_counter() - start_time > timout_s:
            break

    return (time.perf_counter() - start_time) / count * 1000.0


def search_speed_test(
    index: faiss.Index, query: Optional[np.ndarray] = None, ksearch: int = 40, timout_s: Union[float, int] = 10.0
) -> Dict[str, float]:
    """return the average and 99p search speed"""

    nb_samples = 2_000

    if query is None:
        query = np.random.rand(nb_samples, index.d).astype("float32")

    test_start_time_s = time.perf_counter()
    speed_list_ms = []  # in milliseconds

    nb_repeat = 1 + (nb_samples - 1) // query.shape[0]

    for one_query in chain.from_iterable(repeat(query, nb_repeat)):

        start_time_s = time.perf_counter()  # high precision
        _, _ = index.search(np.expand_dims(one_query, 0), ksearch)
        end_time_s = time.perf_counter()

        search_time_ms = 1000.0 * (end_time_s - start_time_s)
        speed_list_ms.append(search_time_ms)

        if time.perf_counter() - test_start_time_s > timout_s:
            break

    speed_list_ms = np.array(speed_list_ms)
    print(len(speed_list_ms))

    # avg2 = 1000 * (time.perf_counter() - test_start_time_s) / len(speed_list_ms)

    speed_infos = {
        "avg_search_speed_ms": np.average(speed_list_ms),
        "99p_search_speed_ms": np.quantile(speed_list_ms, 0.99),
    }

    return speed_infos


def format_speed_ms_per_query(speed: float) -> str:
    """format the speed (ms/query) into a nice string"""
    return f"{speed:.2f} ms/query"


def quantize_vec_without_modifying_index(index: faiss.Index, vecs: np.ndarray) -> np.ndarray:
    """qantize a batch of vectors"""
    quantized_vecs = index.sa_decode(index.sa_encode(vecs))
    return quantized_vecs


def set_search_hyperparameters(index: faiss.Index, param_str: str, use_gpu: bool = False) -> None:
    """set hyperparameters to an index"""
    # depends on installed faiss version # pylint: disable=no-member
    params = faiss.ParameterSpace() if not use_gpu else faiss.GpuParameterSpace()
    params.set_index_parameters(index, param_str)


def save_index_on_hdfs(my_index: faiss.Index, dst_path: str) -> bool:
    """function to save a faiss index on hdfs"""

    index_path = "/tmp/tmp_index.index"

    if os.path.exists(index_path):
        os.remove(index_path)

    faiss.write_index(my_index, index_path)
    cmd = f"hdfs dfs -copyFromLocal {index_path} {dst_path}"

    if not run_command(cmd):
        print("Overwriting previous file")
        cmd = f"hdfs dfs -copyFromLocal -f {index_path} {dst_path}"
        if not run_command(cmd):
            os.remove(index_path)
            return False

    os.remove(index_path)

    return True


def load_index_from_hdfs(src_path: str):
    """Load a faiss index stored on the HDFS"""

    index_path = f"/tmp/{src_path.split('/')[-1]}"

    if os.path.exists(index_path):
        os.remove(index_path)

    cmd = f"hdfs dfs -copyToLocal {src_path} {index_path}"

    if not run_command(cmd):
        raise ValueError(f"no file at {src_path} on HDFS")

    index_memory = os.path.getsize(index_path)

    index = faiss.read_index(index_path)

    os.remove(index_path)

    return index, index_memory



""" This file contains a wrapper class to create Faiss-like indices """

from abc import ABC, abstractmethod

import faiss
import numpy as np


class FaissIndexWrapper(ABC):
    """
    This abstract class is describing a Faiss-like index
    It is useful to use this wrapper to use benchmarking functions written for
    faiss in this library
    """

    # pylint: disable=invalid-name
    def __init__(self, d: int, metric_type: int):
        """
        __init__ function for FaissIndexWrapper

        Parameters:
        -----------
        d : int
            dimension of the vectors, named d to keep Faiss notation
        metric_type : int
            similarity metric used in the vector space, using faiss
            enumerate values (faiss.METRIC_INNER_PRODUCT and faiss.METRIC_L2)
        """

        self.d = d

        if metric_type in [faiss.METRIC_INNER_PRODUCT, "IP", "ip"]:
            self.metric_type = faiss.METRIC_INNER_PRODUCT
        elif metric_type in [faiss.METRIC_L2, "L2", "l2"]:
            self.metric_type = faiss.METRIC_L2
        else:
            raise NotImplementedError

    # pylint: disable=invalid-name
    @abstractmethod
    def search(self, x: np.ndarray, k: int):
        """
        Function that search the k nearest neighbours of a batch of vectors

        Parameters
        ----------
        x : 2D numpy.array of floats
            Batch of vectors of shape (batch_size, vector_dim)
        k : int
            Number of neighbours to retrieve for every vector

        Returns
        -------
        D : 2D numpy.array of floats
            Distances numpy array of shape (batch_size, k).
            Contains the distances computed by the index of the k nearest neighbours.
        I : 2D numpy.array of ints
            Labels numpy array of shape (batch_size, k).
            Contains the vectors' labels of the k nearest neighbours.
        """

        raise NotImplementedError

    # pylint: disable=invalid-name
    @abstractmethod
    def add(self, x: np.ndarray):
        """
        Function that adds vectors to the index

        Parameters
        ----------
        x : 2D numpy.array of floats
            Batch of vectors of shape (batch_size, vector_dim)
        """

        raise NotImplementedError



""" functions that read efficiently files from hdfs without saving it on disk """

from typing import Iterator, List

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from tqdm import tqdm as tq


def read_filenames(parquet_embeddings_path: str) -> List[str]:
    """
    Read the name of all the parquet files

    Parameters
    ----------
    parquet_embeddings_path: str
        Path to .parquet embedding files on hdfs

    Returns
    -------
    filenames: List[str]
        The list of the paths to .parquet files of the given directory sorted in
        non decreasing order.
    """

    hdfs = pa.hdfs.connect()

    all_filenames = hdfs.ls(parquet_embeddings_path)
    filenames = ["hdfs:" + path.split(":")[-1] for path in all_filenames if path.endswith(".parquet")]
    filenames.sort()

    return filenames


def read_embeddings_remote(
    embeddings_path: str, column_label: str = "embedding", stack_input: int = 1, verbose=True
) -> Iterator[np.ndarray]:
    """
    Return an iterator over embeddings from a parquet folder

    Parameters
    ----------
    embeddings_path : str
        Path on the hdfs of the embedding in parquet format.
    column_label : str (default "embeddings")
        Name of the column in which the embeddings are stored.
    stack_input : int (default 1)
        Number of arrays that should be stacked at each iterations.
        This parameter is useful when working with many small files.
    verbose : bool
        Print detailed informations if set to True

    Returns
    -------
    embeddings_iterator : Iterator[np.ndarray]
        An iterator over batchs of embedding arrays.
    """

    assert stack_input > 0

    filenames = read_filenames(embeddings_path)

    embeddings_stack: List[np.ndarray] = []

    iterator = list(enumerate(filenames))
    if verbose:
        iterator = tq(iterator)

    for file_number, file_name in iterator:

        if embeddings_stack and (file_number % stack_input == 0):
            yield np.concatenate(embeddings_stack)
            embeddings_stack = []

        small_table = pq.read_table(file_name)
        pandas_df = small_table[column_label].to_pandas()
        embeddings_stack.append(np.stack(pandas_df).astype("float32"))

    if embeddings_stack:
        yield np.concatenate(embeddings_stack)



""" function to interact with os """

import os
import subprocess
from typing import List


def run_command(cmd: str) -> bool:
    """Function to run a bash command"""
    try:
        subprocess.run(cmd.split(), check=True)
        return True
    except subprocess.CalledProcessError:
        return False


def list_local_files(path: str) -> List[str]:
    """function to list the files in a directory"""

    for infos in os.walk(path):
        return infos[2]
    return []



#!/usr/bin/env python3
"""Setup script"""

from pathlib import Path
import re

import setuptools

if __name__ == "__main__":

    # Read metadata from version.py
    with Path("autofaiss/version.py").open(encoding="utf-8") as file:
        metadata = dict(re.findall(r'__([a-z]+)__\s*=\s*"([^"]+)"', file.read()))

    # Read description from README
    with Path(Path(__file__).parent, "README.md").open(encoding="utf-8") as file:
        long_description = file.read()

    _INSTALL_REQUIRES = [
        "dataclasses",
        "fire>=0.4.0",
        "numpy>=1.18.2",
        "pandas>=1.0.5",
        "pyarrow>=0.14",
        "tqdm>=4.46.0",
        "faiss-cpu>=1.7.0",
    ]

    _TEST_REQUIRE = ["pytest"]

    # Run setup
    setuptools.setup(
        name="autofaiss",
        version=metadata["version"],
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Intended Audience :: Developers",
        ],
        long_description=long_description,
        long_description_content_type="text/markdown",
        description=long_description.split("\n")[0],
        author=metadata["author"],
        install_requires=_INSTALL_REQUIRES,
        tests_require=_TEST_REQUIRE,
        dependency_links=[],
        entry_points={"console_scripts": ["autofaiss = autofaiss.external.quantize:main"]},
        data_files=[(".", ["requirements.txt", "README.md"])],
        packages=setuptools.find_packages(),
        url="https://github.com/criteo/autofaiss",
    )



""" functions to download and convert embedding parquet files """

import os

import numpy as np
from tqdm import tqdm as tq

from autofaiss.datasets.readers.remote_iterators import read_filenames


def save(filename: str, dest_path: str) -> None:
    """
    Function that loads and converts an hdfs embeddings .parquet file to
    numpy arrays and store it in the local desination path.
    """

    basename = filename.split("/")[-1].split(".")[0]

    embeddings_array_path = f"{dest_path}/{basename}_embeddings.npy"

    if not os.path.exists(embeddings_array_path):

        # pylint: disable=import-outside-toplevel
        import pyarrow.parquet as pq

        small_table = pq.read_table(filename, use_threads=False)
        pandas_df = small_table.to_pandas()
        embeddings = np.stack(pandas_df["embedding"].to_numpy()).astype("float32")

        with open(embeddings_array_path, "wb") as embedding_file:
            np.save(embedding_file, embeddings)


def download(parquet_embeddings_path, dest_path, n_cores=1, verbose=True) -> bool:
    """
    Download and convert all the embedding parquet file from the hdfs path
    to numpy arrays.
    Parallelisation is not possible since the connection to hdfs is unique (singleton class)
    """

    os.makedirs(dest_path, exist_ok=True)

    filenames = list(reversed(read_filenames(parquet_embeddings_path)))

    if n_cores != 1:
        raise ValueError("multiprocessing is not compatible")

    if verbose:
        filenames = tq(list(filenames))
    for filename in filenames:
        save(filename, dest_path)

    return True


if __name__ == "__main__":

    DEST_PATH = "/home/v.paltz/downloaded_vectors/image_embeddings_100_views_US"
    SRC_PATH = "/user/deepr/dev/r.beaumont/image_embeddings_100_views_US"

    download(SRC_PATH, DEST_PATH)
